/*******************************************************************
 * MACRO: ScoreXML
 * PURPOSE: generates an XML file to describe the metadata for
 *          the input and output variables in the specified
 *          data step code and input data set.
 * Usage:
 *        %LET _XMLVERNUM= ;
 *        %LET _EMDEBUG = 0;
 *        %LET _VARMETADATA = 0;
 *        Libname INDATA "C:\temp\example";
 *        Filename DSFILE "C:\temp\example\score.sas ";
 *        Filename XMLFILE "C:\temp\example\score.xml";
 *        %ScoreXML(CodeFile= DSFILE
 *                 ,DataLibref= INDATA.Mydata
 *                 ,XMLFilename= XMLFILE
 *                 ,KeepVarsData= INDATA.varnames
 *                 ,JavaPkgName= eminer.user.Score
 *                 ,XMLfmt=GENERIC
 *                 ,XMLProducer=SAS Enterprise Miner);
 *
 * Required parameters:
 * CodeFile - a SAS filename for SAS data step code (no step boundaries).
 * Data - libref.membername data must match data step vars. (extra vars ok)
 * XMLFile - destination of XML as a SAS filename (logical name) for external file.
 *
 * Optional Parameters:
 * XMLfmt -  Default is GENERIC, optional CSCORE XML has a <CFunction> section.
 * XMLProducer -  Default = SAS Enterprise Miner. String used as name in <Producer> section.
 * KeepVarsData - libref.membername for data set with Name for keep vars.
 * JavaPkgName - Character value for the EM generated Java Score class package name.
 *
 * Optional Global macros:
 *  _XMLVERNUM - alternate xml version default is 6.2.
 * _EMDEBUG - set to 1 to prevents the macro from deleting it's work files.
 * _VARMETADATA - set to 1 retains the _Invarsmeta and _Outvarsmeta data sets in work.
 *
 * Sub-macros
 * called:
 *   MkVarXML - writes the XML for both Input and Output elements.
 *
 * Data sets
 * created:
 *   _alldeps     - the output of proc fcmp, all the code variables and
 *                   their dependencies
 *   _indatavmeta - the variables metadata for the input data
 *   _keepvars    - only the variables in both the optional KeepVarsData
 *                   and _Varsmeta data sets
 *   _invarsmeta  - input variables from executed code
 *   _progout     - the output from running the data step code with the
 *                   input data
 *   _outvarsmeta - output variables from executed code
 *   _rhs         - only the code variables on the Right Hand Side
 *   _varsmeta    - metadata for all variables in program output (_Progout)
 *
 * Limitations:
 *   Mismatched code and data are problematic.
 *    Mismached variable names, variable type and other attributes
 *    are not checked and may produce XML that does not reflect the
 *    code accurately.
 *
 ********************************************************************/

%Macro ScoreXML(CodeFile=, Data=, XMLFile=,
         XMLfmt=GENERIC, XMLProducer=SAS Enterprise Miner,
         KeepVarsData=, JavaPkgName=);

   %GLOBAL _EMDEBUG _VARMETADATA _XMLVERNUM;

   %LOCAL argError keepVarsOpt pkgNameOpt codeFilePath rc
          xmlFilePath vcheck xmlVersionNum dsid num_MissVars
          num_InVars num_OutVars vcheckDBG vcheckMeta;

   %let SYSCC = 0;
   %let argError=false;
   %let keepVarsOpt=false;
   %let pkgNameOpt=false;

   %put;

   /*---------------------------------------------------------------
    * Check parameters
    *---------------------------------------------------------------*/
   %if (("&CodeFile"= "") or (%length(%superq(CodeFile))<=0)) %then
   %do;
                               /* CodeFile parameter required   */
      %put ERROR: ScoreXML: CodeFile parameter required;
      %let argError=true;
   %end;
   %else %if (%SYSFUNC(FEXIST(&CodeFile)) ne 1) %then
      %do;
                               /* Code File is not found            */
         %put ERROR: ScoreXML: Code File is not found;
         %let argError=true;
      %end;
   %else %do;
      %let codeFilePath = %NRBQUOTE(%SYSFUNC(pathname(&&CodeFile,F)));
      %put NOTE: ScoreXML: Processing code file = %NRBQUOTE(&codeFilePath);
      %put;
   %end;

   %if (("&Data"="") or (%length(%superq(Data))<=0)) %then
   %do;
                               /* Data parameter required   */
      %put ERROR: ScoreXML: Data argument required;
      %let argError=true;
   %end;
   %else %do;
      %let rc = %SYSFUNC(exist(&Data,DATA));
      %let rcv = %SYSFUNC(exist(&Data,VIEW));
      %if (%eval(&rc ^= 1)) and (%eval(&rcv ^= 1))  %then
      %do;
                               /* DataLibrary not found             */
         %put ERROR: ScoreXML: Data = &Data, not found;
         %put %SYSFUNC(sysmsg());
         %let argError=true;
      %end;
      %else
           %put NOTE: ScoreXML: Processing Data = &Data. ;
   %end;

   %if (("&XMLFile"="") or (%length(&XMLFile)<=0)) %then
   %do;
                               /* XMLFile parameter required    */
      %put ERROR: ScoreXML: XMLFile argument required.;
      %let argError=true;
   %end;
   %else %if (%SYSFUNC(FILEREF(&XMLFile)) > 0) %then
   %do;
                               /* XML Filename is not assigned      */
      %put ERROR: ScoreXML: XMLFile "&XMLFile" is not assigned .;
      %let argError=true;
   %end;
   %else %do;
      %let xmlFilePath = %NRBQUOTE(%SYSFUNC(pathname(&&XMLFile,F)));
      %put NOTE: ScoreXML: Processing XML file = %NRBQUOTE(&xmlFilePath).;
   %end;

   %if (("&XMLfmt"^="CSCORE") and ("&XMLfmt"^="GENERIC")) %then
   %do;
                                  /* invalid XML format specified    */
         %put ERROR: ScoreXML: invalid XML format specified ;
         %let argError=true;
   %end;

    /* If optional Keep vars DataLibrary   */
   %if (("&KeepVarsData" ^= "") and (%length(%superq(KeepVarsData))> 0)) %then
   %do;
      %let rc = %SYSFUNC(exist(&KeepVarsData,DATA));
      %if (%eval(&rc ^= 1)) %then
      %do;
                               /* DataLibrary not found             */
         %put ERROR: ScoreXML: KeepVarsData = &KeepVarsData, not found;
         %put %SYSFUNC(sysmsg());
         %let argError=true;
      %end;
      %else %do;
         %let keepVarsOpt=true;
         %put NOTE: ScoreXML: Processing Keep variables list from Data library = &KeepVarsData. ;
      %end;
   %end;

    /* If optional Java Score Class name specified   */
   %if ( ("&JavaPkgName" ^= "") and (%length(%superq(JavaPkgName))> 0) ) %then
   %do;
         %let pkgNameOpt=true;
         %put NOTE: ScoreXML: Processing Java Score class name = &JavaPkgName. ;
   %end;

   /*---------------------------------------------------------------
    * If error in parameters detected do not continue
    *---------------------------------------------------------------*/
   %if &argError=true %then
   %do;
                               /* Terminate if invalid arguments    */
      %put ERROR: ScoreXML: Invalid arguments;
      %goto errexit;
   %end;

   /*---------------------------------------------------------------
    * Check for alternate xml version
    *---------------------------------------------------------------*/
   %let vcheck=%symexist(_XMLVERNUM);
   %if (&vcheck = 0) or ((&vcheck = 1)and(%length(&_XMLVERNUM)<=0)) %then
   %do;
      %let xmlVersionNum =6.2; /* default xml format version number */
   %end;
   %put NOTE: ScoreXML: XML version = &xmlVersionNum.;
   %put;

   /*---------------------------------------------------------------
    * Call FCMP for dependencies
    *---------------------------------------------------------------*/
   %put;
   %put NOTE: ScoreXML: Checking Code for variable dependencies.;
   PROC FCMP  NOEXECUTE IMPUTEDDEPS OUTDEPS=WORK._alldeps;
      %INCLUDE &CodeFile;
   RUN;
   QUIT;

  %if (%eval(&SYSERR > 4)) %then %do;
      %put %SYSFUNC(sysmsg());
      %PUT ERROR: ScoreXML: Unable to determine Code variable dependencies. SYSERR= &SYSERR;
      %goto errexit;
   %end;

   /* check for empty _alldeps */
   Data _NULL_;
      numobs=0;
      dsid=OPEN("WORK._alldeps","i");
         if dsid then do;
            numobs= attrn(dsid,"nlobs");
         end;
         else do;
            put "ERROR: ScoreXML: Unable to open dependency data.";
            msg= sysmsg();
            put msg;
            call symput('SYSCC','1012');
         end;
         if numobs <= 0 then do;
            put "WARNING: ScoreXML: No variable dependencies found in code.";
            call symput('SYSCC','4');
         end;
         dsid = close(dsid);
   Run;
   %if (%eval(&SYSERR > 4)) %then %do;
       %goto errexit;
   %end;

    /* save from _alldeps rhs */
   Data WORK._Rhs (keep=rhs RHS_NLiteral UNAME);
   Set WORK._alldeps;
       Length UNAME $32;
       LABEL rhs = ;
       LABEL RHS_NLiteral = ;
      if RHS_Dropped = 0 AND (1 NE index(rhs, '_OLD_')) then do;
         UNAME = UPCASE(rhs);
         output;
      end;
   Run;

   %let rc = %SYSFUNC(exist(WORK._Rhs));
      %if (%eval(&rc ^= 1)) %then
      %do;
         %put %SYSFUNC(sysmsg());
         %put ERROR: ScoreXML: Intermediate file creation failed.;
         %let SYSCC = 1012;
         %goto errexit;
      %end;

   /* use proc sort to remove dups and get right hand side vars */
   Proc Sort Data=WORK._Rhs NODUP out=WORK._Rhs (rename=(rhs=NAME RHS_NLiteral=NLITERAL));
      By UNAME;
   Run;
   %if (%eval(&SYSERR > 4)) %then %do;
      %put %SYSFUNC(sysmsg());
      %goto errexit;
   %end;

   /*---------------------------------------------------------------
    * Generate vars metadata for all vars in the input data set
    *---------------------------------------------------------------*/
   Data WORK._Indatavmeta (KEEP= NAME UNAME TYPE LENGTH LABEL FORMAT KEEP);
      LENGTH NAME $ 32;
      LENGTH UNAME $ 32;
      LENGTH TYPE $ 1;
      LENGTH LENGTH $ 8;
      LENGTH LABEL $ 256;
      LENGTH FORMAT $32;

      KEEP = 0;

      dsid=OPEN("&Data","i");
      if dsid then do;
         numvars= attrn(dsid,"nvars");

         do i=1 to numvars;
            NAME = varname(dsid,i);
            UNAME = UPCASE(NAME);
            vtype = vartype(dsid,i);
            /* convert type to character indicator */
            if vtype = 'N' then TYPE = 'N';
            else TYPE = 'C';

            nlen = varlen(dsid,i);
            LENGTH = put(nlen, 8.);

            LABEL = varlabel(dsid,i);
            /* check length of label */
            len = klength(LABEL);
            if (len = 1 or len = 0) and (kindex(LABEL,"'")=1) then
               LABEL = " ";

            FORMAT = varfmt(dsid,i);

            OUTPUT;
          end;
          dsid = close(dsid);
      end;

      else do;
                               /*  failed open score output data    */
         put "ERROR: ScoreXML: Unable to open input data.";
         msg= sysmsg();
         put msg;
         call symput('SYSCC','1012');
      end;
   RUN;

   %if (%eval(&SYSERR > 4)) %then %do;
      %goto errexit;
   %end;

   /* Sort _Indatavmeta for following merge steps */
   Proc Sort Data=Work._Indatavmeta  NODUP Out=Work._Indatavmeta ;
      BY UNAME;
   Run;
   %if (%eval(&SYSERR > 4)) %then %do;
      %put %SYSFUNC(sysmsg());
      %goto errexit;
   %end;

   /*---------------------------------------------------------------
    * Run the data step code against the input data set
    *---------------------------------------------------------------*/
    /* First optimize for Libname engine */
    %let _engineTable=;
    %let _dsidTable  = 0;
     /* Retrieve the libname engine */
    Data _null_;
       dsid = open("&data");
       call symput('_dsidTable', strip(put(dsid, best.)));
       if dsid then do;
          call symput('_engineTable', upcase(attrc(dsid, 'ENGINE')));
          dsid = close(dsid);
       end;
    Run;

    %if "&_engineTable" eq "GREENPLM" %then %do;
        Data WORK._Progout;
           Set &Data(obs=1);
           %include &CodeFile;
       Run;
    %end;
    %else %do;
        Data WORK._Progout;
           Set &Data;
           %include &CodeFile;
           if _N_=1 then do;
              output;
              stop;
           end;
        Run;
   %end;

   %if (%eval(&SYSERR > 4)) %then
   %do;
      %put %SYSFUNC(sysmsg());
      %put ERROR: ScoreXML: Code execution failed.;
      %let SYSCC = &SYSERR;
      %goto errexit;
   %end;
   %else %let SYSCC=0;         /* in case of a warning reset SYSCC  */

   %let rc = %SYSFUNC(exist(WORK._Progout));
   %if (%eval(&rc ^= 1)) %then
      %do;
         %put %SYSFUNC(sysmsg());
         %put ERROR: ScoreXML: Program results file creation failed.;
         %let SYSCC = 1012;
         %goto errexit;
      %end;

   /*---------------------------------------------------------------
    * Generate vars metadata for all vars in program output
    *---------------------------------------------------------------*/
   Data WORK._Varsmeta (KEEP= NAME UNAME TYPE LENGTH LABEL FORMAT KEEP);
      LENGTH NAME $ 32;
      LENGTH UNAME $ 32;
      LENGTH TYPE $ 1;
      LENGTH LENGTH $ 8;
      LENGTH LABEL $ 256;
      LENGTH FORMAT $32;

      KEEP = 0;

      dsid=OPEN("WORK._Progout","i");
      if dsid then do;
         numvars= attrn(dsid,"nvars");

         do i=1 to numvars;
            Name = varname(dsid,i);
            UNAME = UPCASE(NAME);
            vtype = vartype(dsid,i);
            /* convert type to character indicator */
            if vtype = 'N' then TYPE = 'N';
            else TYPE = 'C';

            nlen = varlen(dsid,i);
            LENGTH = put(nlen, 8.);

            LABEL = varlabel(dsid,i);
           /* check length of label */
            len = klength(LABEL);
            if (len = 1 or len = 0) and (kindex(LABEL,"'")=1) then
               LABEL = " ";

            FORMAT = varfmt(dsid,i);

            OUTPUT;
          end;
          dsid = close(dsid);
      end;

      else do;
                               /*  failed open score output data    */
         put "ERROR: ScoreXML: Unable to open program output.";
         msg= sysmsg();
         put msg;
         call symput('SYSCC','1012');
      end;
   RUN;

   %if (%eval(&SYSERR > 4)) %then %do;
      %goto errexit;
   %end;

   /* Sort _Varsmeta for following merge steps */
   Proc Sort Data=Work._Varsmeta  NODUP Out=Work._Varsmeta ;
      BY UNAME;
   Run;
   %if (%eval(&SYSERR > 4)) %then %do;
      %put %SYSFUNC(sysmsg());
      %goto errexit;
   %end;

   /*---------------------------------------------------------------
    * Process the Keep vars data to add keep flag to _Varsmeta
    *---------------------------------------------------------------*/
   %if &keepVarsOpt=true %then
   %do;
      Data WORK._Keepnames;
      Set &KeepVarsData;
      LENGTH UNAME $ 32;
         UNAME = UPCASE(NAME);
      Run;
      Proc Sort Data=WORK._Keepnames NODUP out=WORK._Keepvars ;
         By UNAME;
      Run;
      %if (%eval(&SYSERR > 4)) %then %do;
         %put %SYSFUNC(sysmsg());
         %goto errexit;
      %end;

     Data WORK._Varsmeta (KEEP= NAME UNAME TYPE LENGTH LABEL FORMAT KEEP);
        MERGE    WORK._Varsmeta (in=a)  WORK._Keepvars (in=b);
        BY UNAME;
         if a and b then KEEP = 1;
         If a then output;
      Run;
      %if (%eval(&SYSERR > 4)) %then
      %do;
         %put %SYSFUNC(sysmsg());
         DATA _NULL_;
         SET WORK._Keepvars (obs=1);
            if length(NAME) ^= 32 then
               put "ERROR: ScoreXML: KeepVarsData variable NAME must be length 32.";
         RUN;
         %goto errexit;
      %end;
   %end;

   /*---------------------------------------------------------------
    * Determine program inputs
    * program input vars must be used on rhs and not on lhs
    *---------------------------------------------------------------*/
   Data WORK._Invarsmeta;
      MERGE  WORK._Rhs (in=a) WORK._Indatavmeta (in=b);
      BY UNAME;
      if (a and b)  then output;
   Run;
   %if (%eval(&SYSERR > 4)) %then
   %do;
      put "ERROR: ScoreXML: Unable to determine program input variables.";
      %goto errexit;
   %end;

   /*---------------------------------------------------------------
    * Determine program outputs
    * prog outputs are not program inputs
    *---------------------------------------------------------------*/
   Data WORK._Outvarsmeta;
      MERGE    WORK._Varsmeta (in=a)  WORK._Indatavmeta (in=b);
      BY UNAME;
         If a and not b then output;
   Run;
   %if (%eval(&SYSERR > 4)) %then
   %do;
      %put ERROR: ScoreXML: Unable to determine program output variables.;
      %goto errexit;
   %end;

   /* check the number of Input vars */
   %let num_InVars = 0;
   %let dsid = %sysfunc(open(WORK._Invarsmeta));
   %if &dsid %then %do;
       %let num_InVars = %sysfunc(attrn(&dsid, NLOBS));
       %let dsid = %sysfunc(close(&dsid));
   %end;
   %if (%eval(&num_InVars <= 0)) %then
   %do;
      %PUT WARNING: ScoreXML: No inputs detected.;
   %end;

   /* check the number of Output vars */
   %let num_OutVars = 0;
   %let dsid = %sysfunc(open(WORK._Outvarsmeta));
   %if &dsid %then %do;
       %let num_OutVars = %sysfunc(attrn(&dsid, NLOBS));
       %let dsid = %sysfunc(close(&dsid));
   %end;
   %if (%eval(&num_OutVars <= 0)) %then
   %do;
      %PUT WARNING: ScoreXML: No outputs detected.;
   %end;

   /*---------------------------------------------------------------
    * Start writing XML file
    *---------------------------------------------------------------*/
   FILENAME XMLfile "&xmlFilePath" encoding="UTF-8" NOBOM;

   Data _null_;
      LENGTH strbuf $ 80;
      strbuf = ' ';
      file XMLfile;

      put '<?xml version="1.0" encoding="utf-8"?>';
      put '<Score>';
      put @3 '<Producer>';
      put @6 "<Name> &XMLProducer </Name>";
      strbuf = '<Version> ' || "&xmlVersionNum" || ' </Version>';
      put @6 strbuf;
                               /* get POSIX style locale            */
      syslocale = getpxlocale();
      strbuf = '<Locale> '|| strip(syslocale) ||' </Locale>';
      put @6 strbuf;
                               /* get SAS 12 byte encoding name     */
      sysencode = getpxencoding('sas12');
      strbuf = '<Encoding> '|| strip(sysencode) ||' </Encoding>';
            put @6 strbuf;

      put @3 '</Producer>';

     %if &pkgNameOpt=true %then
     %do;
        strbuf = '<JavaPkgName> ' || "&JavaPkgName" || ' </JavaPkgName>';
        put @3 strbuf;
     %end;

      if "&XMLfmt" = "CSCORE" then do;
         put @3 '<CFunction>';
         put @6 '<Name> score </Name>';
         put @6 '<Parameter usage="input">';
         put @9 '<Type> Parm* </Type>';
         strbuf = '<NumElements> ' || "&num_InVars" || ' </NumElements>';
         put @9 strbuf;
         put @6 '</Parameter>';
         put @6 '<Parameter usage="output">';
         put @9 '<Type> Parm* </Type>';
         strbuf = '<NumElements> ' || "&num_OutVars" || ' </NumElements>';
         put @9 strbuf;
         put @6 '</Parameter>';
         put @3 '</CFunction>';
      end;
   Run;

   /* append the xml for both inputs and outputs */
   %MkVarXML(WORK._Invarsmeta, INPUT, %NRBQUOTE(&xmlFilePath), OptFmtFlag=&XMLfmt, OptAppend='Y', OptKeepVars= &keepVarsOpt);
    %if (%eval(&SYSERR > 4)) %then %do;
      %PUT ERROR: ScoreXML: Writing input variable xml failed. SYSCC= &SYSCC;
      %goto errexit;
   %end;

   %MkVarXML(WORK._Outvarsmeta, OUTPUT, %NRBQUOTE(&xmlFilePath), OptFmtFlag=&XMLfmt, OptAppend='Y', OptKeepVars= &keepVarsOpt);
    %if (%eval(&SYSERR > 4)) %then %do;
      %PUT ERROR: ScoreXML: Writing output variable xml failed. SYSCC= &SYSCC;
      %goto errexit;
   %end;

   /* close the XML Score element with a tag */
   FILENAME XMLfile "&xmlFilePath" encoding="UTF-8" NOBOM MOD;
   Data _null_;
   file XMLfile;
      put '</Score>';
   Run;

%goto clean;
%errExit:

   %if (%eval(&SYSERR > 4) or %eval(&SYSCC > 4)) %then %do;

      %let SYSCC = 8;
      %end;
   %put ERROR: ScoreXML: Terminating because of previous errors. SYSCC= &SYSCC;
   %put;

%clean:
   /* Clean up */
   /* if _EMDEBUG is not = 1  delete all macro created data sets */
   %let vcheckDBG=%symexist(_EMDEBUG);
   %if (&vcheckDBG = 0) or ((&vcheckDBG = 1)and("&_EMDEBUG" ^= "1")) %then
   %do;
      /* if _VARMETADATA ^= 1 delete all else retain _Invarsmeta and _Outvarsmeta */
      %let vcheckMeta=%symexist(_VARMETADATA);
      %if (&vcheckMeta = 0) or ((&vcheckMeta = 1)and("&_VARMETADATA" ^= "1")) %then
      %do;
         /* delete all macro created data sets */
         Proc Datasets library=WORK NOLIST NOWARN NODETAILS;
         DELETE _alldeps _Indatavmeta _Keepvars _Keepnames _Invarsmeta _Progout _Outvarsmeta _Rhs _Varsmeta ;
         Run;
         Quit;
      %end;
      %else
      %do;
         /* leave _Invarsmeta and _Outvarsmeta */
         Proc Datasets library=WORK NOLIST NOWARN NODETAILS;
         DELETE _alldeps _Indatavmeta _Keepvars _Keepnames _Progout  _Rhs _Varsmeta ;
         Run;
         Quit;
      %end;
   %end;

   %RETURN;
%Mend ScoreXML;

/*******************************************************************
 * MkVarXML a subroutine for the ScoreXML macro
 * Used to write xml for both input or output type vars
 *
 ********************************************************************/
%Macro MkVarXML(VarMeta, TypeVars, XmlPath, OptFmtFlag= GENERIC, OptAppend='Y', OptKeepVars= false );
   %LOCAL num_vars;
   %LOCAL dsid;

   /* get the number of vars described by the metadata */
   %let num_vars = 0;
   %let dsid = %sysfunc(open(&varmeta));
   %if &dsid %then %do;
       %let num_vars = %sysfunc(attrn(&dsid, NLOBS));
       %let dsid = %sysfunc(close(&dsid));
   %end;
   %else %do;
      %put %SYSFUNC(sysmsg());
      %return;
   %end;

   %if &OptAppend = 'Y' %then %do;
      FILENAME XMLfile "&XmlPath" encoding="UTF-8" NOBOM MOD;
      %end;
   %else %do;
      FILENAME XMLfile "&XmlPath" encoding="UTF-8" NOBOM;
   %end;

   Data _null_;
      Set &VarMeta end=eof;
      Retain  varindex 0;
      Length keepflag $5;
      Length strbuf $ 1024;
      Length vindexc $12;
      strbuf = ' ';
      file XMLfile;

      keepflag = "&OptKeepVars";

      vindexc = put(varindex, BEST12.);

      if _N_ = 1 then          /* befor starting first var          */
      do;
                               /* select start tag for in or out    */
         if "&TypeVars" = "INPUT" then
            strbuf =  ' <Input count="' || "&num_vars" || '">';
         else
            strbuf =  ' <Output count="' || "&num_vars" || '">';

          put @3 strbuf;
      end;

      if keepflag = "true" then do;
         if (KEEP = 1) then
            strbuf =  '<Variable index="'||strip(vindexc)||'" keep="true">';
                     else
            strbuf =  '<Variable index="'||strip(vindexc)||'" keep="false">';
         end;
      else
         strbuf =  '<Variable index="'||strip(vindexc)||'">';
      put @6 strbuf;

      /* if validvarname=any variable name */
      if (NLITERAL = 1) then do;
         strbuf = '<Vname>';
         put @9 strbuf;
         strbuf = '<![CDATA['||trim(name)||']]>';
         put @12 strbuf;
         put @9'</Vname>';
      end;
      else do;
         strbuf = '<Name> '||strip(name)||' </Name>';
         put @9 strbuf;
      end;

      /* get var type */
      if "&OptFmtFlag" = "CSCORE" then do;
         CtypeVal = 'char';
         NtypeVal = 'double';
      end;
      else do;
         CtypeVal = 'C';
         NtypeVal = 'N';
      end;
      if strip(type) = 'C' then
         strbuf = '<Type length="'|| strip(length)||'"> '||strip(CtypeVal)||' </Type>';
      else
         strbuf = '<Type> '||strip(NtypeVal)||' </Type>';

      put @9 strbuf;

      if LengthN(LABEL) then do;
         put @9 '<Description>';
         strbuf = '<![CDATA['||strip(LABEL)||']]>';
         put @12 strbuf;
         put @9 '</Description>';
      end;

      put @6 '</Variable>';

      varindex = varindex +1;

       /* put last */
      if eof then do;
         if "&TypeVars" = "INPUT" then
            strbuf =  "</Input>";
         else
            strbuf =  "</Output>";
         put @3 strbuf;
      end;
   Run;
%Mend MkVarXML;
